

class FeatureSettings:
    def __init__(
            self,
            service_id: int,
            engine_settings: str = None,
            **kwargs
    ):
        self.engine_settings = engine_settings
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




"""
Tests Nitrado connections. If any fail, then this means any of the following:
    - something is wrong with Nitrado
    - the URLs are no longer valid
    - responses from Nitrado has changed
"""
from nitrado import Global
from nitrado.lib import Client




def test_health_check():
    health = Global.health_check()
    error_message = f"Nitrado's API is down or the url is invalid."
    assert health.success, error_message


def test_maintenance():
    maintenance = Global.maintenance_status()
    error_message = f"Nitrado's API is down or the url is invalid."
    assert maintenance.dns_backend is not None, error_message
    assert maintenance.cloud_backend is not None, error_message
    assert maintenance.domain_backend is not None, error_message


def test_version():
    version = Global.version()
    error_message = f"Nitrado's API is down or the url is invalid."
    assert version is not None, error_message





class GameFeatures:
    def __init__(
            self,
            service_id: int,
            has_backups: bool = False,
            has_world_backups: bool = False,
            has_rcon: bool = False,
            has_application_server: bool = False,
            has_container_websocket: bool = False,
            has_file_browser: bool = False,
            has_ftp: bool = False,
            has_expert_mode: bool = False,
            has_packages: bool = False,
            has_plugin_system: bool = False,
            has_restart_message_support: bool = False,
            has_database: bool = False,
            has_playermanagement_feature: bool = False,
            has_curseforge_workshop: bool = False,
            **kwargs
    ):
        self.has_backups = has_backups
        self.has_world_backups = has_world_backups
        self.has_rcon = has_rcon
        self.has_application_server = has_application_server
        self.has_container_websocket = has_container_websocket
        self.has_file_browser = has_file_browser
        self.has_ftp = has_ftp
        self.has_expert_mode = has_expert_mode
        self.has_packages = has_packages
        self.has_plugin_system = has_plugin_system
        self.has_restart_message_support = has_restart_message_support
        self.has_database = has_database
        self.has_playermanagement_feature = has_playermanagement_feature
        self.has_curseforge_workshop = has_curseforge_workshop
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




from __future__ import annotations
from ..lib import Client
from .maintenance import Maintenance
from .game import Game
from .token import Token
from dotenv import load_dotenv, dotenv_values
import os


load_dotenv()


class Global:
    @classmethod
    def health_check(cls) -> Global:
        """Health information about the Nitrado API."""
        response = Client.get_without_api_key('/ping')
        kwargs: dict = response.json()
        return cls(**kwargs)

    @classmethod
    def maintenance_status(cls) -> Maintenance:
        response = Client.get_without_api_key('/maintenance')
        kwargs: dict = response.json()
        return Maintenance(**kwargs['data']['maintenance'])

    @classmethod
    def version(cls) -> str:
        """Nitrado API version."""
        response = Client.get_without_api_key('/version')
        kwargs: dict = response.json()
        return cls(**kwargs).message

    @classmethod
    def full_game_list(cls) -> list[Game]:
        """Get all games supported by Nitrado."""
        path = '/gameserver/games'
        response = Client.get_without_api_key(path=path)
        data: dict = response.json()['data']
        return Game.from_data(**data['games'])

    @classmethod
    def token_info(cls) -> Token:
        path = '/token'
        if Client.ENV_NAME not in dotenv_values() and Client.ENV_NAME not in os.environ:
            return Token()
        response = Client.get(path=path)
        kwargs: dict = response.json()
        if 'data' in kwargs and 'status' in kwargs and kwargs['status'] == 'success':
            return Token.from_data(kwargs['data'])
        return Token()

    def __init__(self, data: dict = None, message: str = None, status: str = None):
        self.success = status == 'success'
        self.data = data
        self.message = message
        self.status = status

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from __future__ import annotations
from ...lib import Client
from ...gameserver import GameServer
from ...gameserver import Players
from .query import Query
from .settings import Settings
from .game_specific import GameSpecific
from ...lib import assert_response_is_ok
from ...lib import assert_response_is_json
import requests


class ArkServer:
    @classmethod
    def unofficial_server_list(cls) -> dict:
        """List of all the unofficial servers."""
        response = requests.get("http://arkdedicated.com/xbox/cache/unofficialserverlist.json")
        assert_response_is_ok(response)
        assert_response_is_json(response)
        return response.json()

    @classmethod
    def official_server_list(cls) -> dict:
        """List of all official servers."""
        response = requests.get("http://arkdedicated.com/xbox/cache/officialserverlist.json")
        assert_response_is_ok(response)
        assert_response_is_json(response)
        return response.json()

    @classmethod
    def banned_list(cls) -> list:
        """List of XUID from all banned players."""
        response = requests.get("http://arkdedicated.com/xboxbanlist.txt")
        assert_response_is_ok(response)
        return response.text.split('\r\n')

    @classmethod
    def find_by_id(cls, service_id: int) -> ArkServer:
        """Find an ArkServer by service id."""
        gameserver = GameServer.find_by_id(service_id)
        if gameserver.game != 'arkxb':
            raise Exception(f"Server ID {service_id} is not from Ark: Survival Evolved (Xbox)")
        data: dict = dict(gameserver)
        data['query'] = Query(service_id, **data['query'])
        data['settings'] = Settings.from_data(service_id, **data['settings'])
        data['game_specific'] = GameSpecific.from_data(service_id, **data['game_specific'])
        return ArkServer(gameserver, **data)

    @classmethod
    def all(cls) -> list[ArkServer]:
        """Get all ArkServers."""
        gameservers = []
        for gameserver in GameServer.all():
            if gameserver.game != 'arkxb':
                continue
            data: dict = dict(gameserver)
            data['query'] = Query(gameserver.service_id, **data['query'])
            data['settings'] = Settings.from_data(gameserver.service_id, **data['settings'])
            data['game_specific'] = GameSpecific.from_data(gameserver.service_id, **data['game_specific'])
            gameservers.append(ArkServer(gameserver, **data))
        return gameservers

    def __init__(
            self,
            gameserver: GameServer,
            query: Query = None,
            settings: Settings = None,
            game_specific: GameSpecific = None,
            **kwargs
    ):
        self.query = query
        self.settings = settings
        self.game_specific = game_specific
        self.service_id = gameserver.service_id
        self.username = gameserver.username
        self.status = gameserver.status
        self.__gameserver = gameserver
        for k, v in kwargs.items():
            self.__dict__[k] = v

    @property
    def map(self) -> str:
        """The map name."""
        return self.query.map

    @property
    def player_max(self) -> int:
        """Maximum players a server can have."""
        return self.query.player_max

    @property
    def player_current(self) -> int:
        """Number of players in the server."""
        return self.query.player_current

    @property
    def admin_password(self) -> str:
        return self.settings.config.admin_password

    @property
    def server_password(self) -> str:
        return self.settings.config.server_password

    @property
    def spectator_password(self) -> str:
        return self.settings.config.spectatorpassword

    @property
    def current_admin_password(self) -> str:
        return self.settings.config.current_admin_password

    def log_shooter_game(self) -> str:
        """The server's logs. Refreshes about every 15+/- minutes """
        path = f'/services/{self.service_id}/gameservers/file_server/download'
        params = {'file': f"/games/{self.username}/noftp/arkxb/ShooterGame/Saved/Logs/ShooterGame.log"}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        url = data['token']['url']
        log_response = requests.get(url)
        assert_response_is_ok(log_response)
        return log_response.text.replace("\r\n", "\n")

    def log_shooter_game_last(self) -> str:
        """The previous server's logs. Refreshes about every 15+/- minutes """
        path = f'/services/{self.service_id}/gameservers/file_server/download'
        params = {'file': f"/games/{self.username}/noftp/arkxb/ShooterGame/Saved/Logs/ShooterGame_Last.log"}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        url = data['token']['url']
        log_response = requests.get(url)
        assert_response_is_ok(log_response)
        return log_response.text.replace("\r\n", "\n")

    def log_restart(self) -> str:
        """The server's system logs. Refreshes about every 15+/- minutes """
        path = f'/services/{self.service_id}/gameservers/file_server/download'
        params = {'file': f"/games/{self.username}/ftproot/restart.log"}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        url = data['token']['url']
        log_response = requests.get(url)
        assert_response_is_ok(log_response)
        return log_response.text.replace("\r\n", "\n")

    def cluster_id(self) -> str:
        """The cluster id of the server."""
        path = f'/services/{self.service_id}/gameservers/games/arkse/gen_cluster_id'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['clusterid']

    def players(self) -> list[Players]:
        return self.__gameserver.players()

    def start(self) -> bool:
        """Starts the Ark server."""
        return self.__gameserver.start_game('arkxb')

    def restart(self, restart_message: str = None, log_message: str = None) -> bool:
        """Restarts the Ark server."""
        return self.__gameserver.restart_game(restart_message=restart_message, log_message=log_message)

    def reinstall(self) -> bool:
        """Reinstall the Ark server."""
        return self.__gameserver.install_game('arkxb', modpack=None)

    def stop(self, message: str = None, stop_message: str = None) -> bool:
        """Stops the Ark server."""
        return self.__gameserver.stop_game(message=message, stop_message=stop_message)

    def uninstall(self) -> bool:
        """Uninstall the Ark server."""
        return self.__gameserver.uninstall_game('arkxb')

    def __repr__(self):
        service_id = f"service_id={repr(self.service_id)}"
        server_name = f"server_name={repr(self.settings.config.server_name)}"
        player_current = f"player_current={repr(self.player_current)}"
        status = f"status={repr(self.status)}"
        params = ", ".join([service_id, server_name, player_current, status])
        return f"<ArkSurvival({params}, ...)>"





from .client import Client
from .errors import assert_response_is_ok
from .errors import assert_response_is_json
from .errors import assert_success
from .errors import assert_response_is_nitrado_data





class GameIni:
    @classmethod
    def from_data(cls, service_id: int, **kwargs):
        return cls(service_id, **{k.lower().replace('-', '_'): v for k, v in kwargs.items()})

    def __init__(
            self,
            service_id: int,
            dinospawnweightmultipliers: str = None,
            overrideengramentries: str = None,
            levelexperiencerampoverrides: str = None,
            overrideplayerlevelengrampoints: str = None,
            tameddinoclassdamagemultipliers: str = None,
            tameddinoclassresistancemultipliers: str = None,
            excludeitemindices: str = None,
            harvestresourceitemamountclassmultipliers: str = None,
            overridenamedengramentries: str = None,
            bonlyallowspecifiedengrams: str = None,
            globalspoilingtimemultiplier: str = None,
            globalitemdecompositiontimemultiplier: str = None,
            globalcorpsedecompositiontimemultiplier: str = None,
            overridemaxexperiencepointsplayer: str = None,
            overridemaxexperiencepointsdino: str = None,
            pvpzonestructuredamagemultiplier: str = None,
            bpvedisablefriendlyfire: str = None,
            resourcenoreplenishradiusplayers: str = None,
            resourcenoreplenishradiusstructures: str = None,
            bautopvetimer: str = None,
            bautopveusesystemtime: str = None,
            autopvestarttimeseconds: str = None,
            autopvestoptimeseconds: str = None,
            layeggintervalmultiplier: str = None,
            dinoturretdamagemultiplier: str = None,
            bdisablelootcrates: str = None,
            dinoharvestingdamagemultiplier: str = None,
            bdisablefriendlyfire: str = None,
            customrecipeeffectivenessmultiplier: str = None,
            customrecipeskillmultiplier: str = None,
            matingintervalmultiplier: str = None,
            egghatchspeedmultiplier: str = None,
            babymaturespeedmultiplier: str = None,
            bpassivedefensesdamageriderlessdinos: str = None,
            killxpmultiplier: str = None,
            harvestxpmultiplier: str = None,
            craftxpmultiplier: str = None,
            genericxpmultiplier: str = None,
            specialxpmultiplier: str = None,
            pgmapname: str = None,
            pgterrainpropertiesstring: str = None,
            configoverridesupplycrateitems: str = None,
            bdisabledinoriding: str = None,
            bdisabledinotaming: str = None,
            busecorpselocator: str = None,
            bdisablestructureplacementcollision: str = None,
            fastdecayinterval: str = None,
            busesingleplayersettings: str = None,
            ballowunlimitedrespecs: str = None,
            supplycratelootqualitymultiplier: str = None,
            fishinglootqualitymultiplier: str = None,
            perlevelstatsmultiplier: str = None,
            babycuddleintervalmultiplier: str = None,
            babycuddlegraceperiodmultiplier: str = None,
            babycuddleloseimprintqualityspeedmultiplier: str = None,
            babyimprintingstatscalemultiplier: str = None,
            playerharvestingdamagemultiplier: str = None,
            cropgrowthspeedmultiplier: str = None,
            babyfoodconsumptionspeedmultiplier: str = None,
            dinoclassdamagemultipliers: str = None,
            bpveallowtribewar: str = None,
            bpveallowtribewarcancel: str = None,
            cropdecayspeedmultiplier: str = None,
            hairgrowthspeedmultiplier: str = None,
            fuelconsumptionintervalmultiplier: str = None,
            kickidleplayersperiod: str = None,
            maxnumberofplayersintribe: str = None,
            usecorpselifespanmultiplier: str = None,
            globalpoweredbatterydurabilitydecreasepersecond: str = None,
            blimitturretsinrange: str = None,
            limitturretsrange: str = None,
            limitturretsnum: str = None,
            bhardlimitturretsinrange: str = None,
            bshowcreativemode: str = None,
            preventofflinepvpconnectioninvincibleinterval: str = None,
            tameddinocharacterfooddrainmultiplier: str = None,
            wilddinocharacterfooddrainmultiplier: str = None,
            wilddinotorpordrainmultiplier: str = None,
            passivetameintervalmultiplier: str = None,
            tameddinotorpordrainmultiplier: str = None,
            bignorestructurespreventionvolumes: str = None,
            bgenesisusestructurespreventionvolumes: str = None,
            bdisablegenesismissions: str = None,
            babyimprintamountmultiplier: str = None,
            hexagonrewardmultiplier: str = None,
            ballowflyerspeedleveling: str = None,
            preventtransferforclassnames: str = None,
            **kwargs
    ):
        self.dinospawnweightmultipliers = dinospawnweightmultipliers
        self.overrideengramentries = overrideengramentries
        self.levelexperiencerampoverrides = levelexperiencerampoverrides
        self.overrideplayerlevelengrampoints = overrideplayerlevelengrampoints
        self.tameddinoclassdamagemultipliers = tameddinoclassdamagemultipliers
        self.tameddinoclassresistancemultipliers = tameddinoclassresistancemultipliers
        self.excludeitemindices = excludeitemindices
        self.harvestresourceitemamountclassmultipliers = harvestresourceitemamountclassmultipliers
        self.overridenamedengramentries = overridenamedengramentries
        self.bonlyallowspecifiedengrams = bonlyallowspecifiedengrams
        self.globalspoilingtimemultiplier = globalspoilingtimemultiplier
        self.globalitemdecompositiontimemultiplier = globalitemdecompositiontimemultiplier
        self.globalcorpsedecompositiontimemultiplier = globalcorpsedecompositiontimemultiplier
        self.overridemaxexperiencepointsplayer = overridemaxexperiencepointsplayer
        self.overridemaxexperiencepointsdino = overridemaxexperiencepointsdino
        self.pvpzonestructuredamagemultiplier = pvpzonestructuredamagemultiplier
        self.bpvedisablefriendlyfire = bpvedisablefriendlyfire
        self.resourcenoreplenishradiusplayers = resourcenoreplenishradiusplayers
        self.resourcenoreplenishradiusstructures = resourcenoreplenishradiusstructures
        self.bautopvetimer = bautopvetimer
        self.bautopveusesystemtime = bautopveusesystemtime
        self.autopvestarttimeseconds = autopvestarttimeseconds
        self.autopvestoptimeseconds = autopvestoptimeseconds
        self.layeggintervalmultiplier = layeggintervalmultiplier
        self.dinoturretdamagemultiplier = dinoturretdamagemultiplier
        self.bdisablelootcrates = bdisablelootcrates
        self.dinoharvestingdamagemultiplier = dinoharvestingdamagemultiplier
        self.bdisablefriendlyfire = bdisablefriendlyfire
        self.customrecipeeffectivenessmultiplier = customrecipeeffectivenessmultiplier
        self.customrecipeskillmultiplier = customrecipeskillmultiplier
        self.matingintervalmultiplier = matingintervalmultiplier
        self.egghatchspeedmultiplier = egghatchspeedmultiplier
        self.babymaturespeedmultiplier = babymaturespeedmultiplier
        self.bpassivedefensesdamageriderlessdinos = bpassivedefensesdamageriderlessdinos
        self.killxpmultiplier = killxpmultiplier
        self.harvestxpmultiplier = harvestxpmultiplier
        self.craftxpmultiplier = craftxpmultiplier
        self.genericxpmultiplier = genericxpmultiplier
        self.specialxpmultiplier = specialxpmultiplier
        self.pgmapname = pgmapname
        self.pgterrainpropertiesstring = pgterrainpropertiesstring
        self.configoverridesupplycrateitems = configoverridesupplycrateitems
        self.bdisabledinoriding = bdisabledinoriding
        self.bdisabledinotaming = bdisabledinotaming
        self.busecorpselocator = busecorpselocator
        self.bdisablestructureplacementcollision = bdisablestructureplacementcollision
        self.fastdecayinterval = fastdecayinterval
        self.busesingleplayersettings = busesingleplayersettings
        self.ballowunlimitedrespecs = ballowunlimitedrespecs
        self.supplycratelootqualitymultiplier = supplycratelootqualitymultiplier
        self.fishinglootqualitymultiplier = fishinglootqualitymultiplier
        self.perlevelstatsmultiplier = perlevelstatsmultiplier
        self.babycuddleintervalmultiplier = babycuddleintervalmultiplier
        self.babycuddlegraceperiodmultiplier = babycuddlegraceperiodmultiplier
        self.babycuddleloseimprintqualityspeedmultiplier = babycuddleloseimprintqualityspeedmultiplier
        self.babyimprintingstatscalemultiplier = babyimprintingstatscalemultiplier
        self.playerharvestingdamagemultiplier = playerharvestingdamagemultiplier
        self.cropgrowthspeedmultiplier = cropgrowthspeedmultiplier
        self.babyfoodconsumptionspeedmultiplier = babyfoodconsumptionspeedmultiplier
        self.dinoclassdamagemultipliers = dinoclassdamagemultipliers
        self.bpveallowtribewar = bpveallowtribewar
        self.bpveallowtribewarcancel = bpveallowtribewarcancel
        self.cropdecayspeedmultiplier = cropdecayspeedmultiplier
        self.hairgrowthspeedmultiplier = hairgrowthspeedmultiplier
        self.fuelconsumptionintervalmultiplier = fuelconsumptionintervalmultiplier
        self.kickidleplayersperiod = kickidleplayersperiod
        self.maxnumberofplayersintribe = maxnumberofplayersintribe
        self.usecorpselifespanmultiplier = usecorpselifespanmultiplier
        self.globalpoweredbatterydurabilitydecreasepersecond = globalpoweredbatterydurabilitydecreasepersecond
        self.blimitturretsinrange = blimitturretsinrange
        self.limitturretsrange = limitturretsrange
        self.limitturretsnum = limitturretsnum
        self.bhardlimitturretsinrange = bhardlimitturretsinrange
        self.bshowcreativemode = bshowcreativemode
        self.preventofflinepvpconnectioninvincibleinterval = preventofflinepvpconnectioninvincibleinterval
        self.tameddinocharacterfooddrainmultiplier = tameddinocharacterfooddrainmultiplier
        self.wilddinocharacterfooddrainmultiplier = wilddinocharacterfooddrainmultiplier
        self.wilddinotorpordrainmultiplier = wilddinotorpordrainmultiplier
        self.passivetameintervalmultiplier = passivetameintervalmultiplier
        self.tameddinotorpordrainmultiplier = tameddinotorpordrainmultiplier
        self.bignorestructurespreventionvolumes = bignorestructurespreventionvolumes
        self.bgenesisusestructurespreventionvolumes = bgenesisusestructurespreventionvolumes
        self.bdisablegenesismissions = bdisablegenesismissions
        self.babyimprintamountmultiplier = babyimprintamountmultiplier
        self.hexagonrewardmultiplier = hexagonrewardmultiplier
        self.ballowflyerspeedleveling = ballowflyerspeedleveling
        self.preventtransferforclassnames = preventtransferforclassnames
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"





class StartParam:
    @classmethod
    def from_data(cls, service_id: int, **kwargs):
        return cls(service_id, **{k.lower().replace('-', '_'): v for k, v in kwargs.items()})

    def __init__(
            self,
            service_id: int,
            PvPStructureDecay: str = None,
            PreventOfflinePvP: str = None,
            PreventOfflinePvPInterval: str = None,
            ShowFloatingDamageText: str = None,
            DisableImprintDinoBuff: str = None,
            AllowAnyoneBabyImprintCuddle: str = None,
            OverideStructurePlatformPrevention: str = None,
            EnableExtraStructurePreventionVolumes: str = None,
            NonPermanentDiseases: str = None,
            PreventDiseases: str = None,
            OverrideStructurePlatformPrevention: str = None,
            PreventTribeAlliances: str = None,
            AllowRaidDinoFeeding: str = None,
            AllowHitMarkers: str = None,
            FastDecayUnsnappedCoreStructures: str = None,
            TribeLogDestroyedEnemyStructures: str = None,
            OverrideOfficialDifficulty: str = None,
            PreventDownloadSurvivors: str = None,
            PreventDownloadItems: str = None,
            PreventDownloadDinos: str = None,
            PreventUploadSurvivors: str = None,
            PreventUploadItems: str = None,
            PreventUploadDinos: str = None,
            ForceFlyerExplosives: str = None,
            DestroyUnconnectedWaterPipes: str = None,
            PvPDinoDecay: str = None,
            PvEAllowStructuresAtSupplyDrops: str = None,
            AllowCrateSpawnsOnTopOfStructures: str = None,
            UseOptimizedHarvestingHealth: str = None,
            ClampItemSpoilingTimes: str = None,
            AutoDestroyDecayedDinos: str = None,
            AllowFlyingStaminaRecovery: str = None,
            AllowMultipleAttachedC4: str = None,
            bAllowPlatformSaddleMultiFloors: str = None,
            PreventSpawnAnimations: str = None,
            AutoDestroyStructures: str = None,
            MinimumDinoReuploadInterval: str = None,
            OnlyAutoDestroyCoreStructures: str = None,
            OxygenSwimSpeedStatMultiplier: str = None,
            ServerAutoForceRespawnWildDinosInterval: str = None,
            CrossARKAllowForeignDinoDownloads: str = None,
            ClampItemStats: str = None,
            EnableCryopodNerf: str = None,
            NewYear1UTC: str = None,
            NewYear2UTC: str = None,
            **kwargs
    ):
        self.PvPStructureDecay = PvPStructureDecay
        self.PreventOfflinePvP = PreventOfflinePvP
        self.PreventOfflinePvPInterval = PreventOfflinePvPInterval
        self.ShowFloatingDamageText = ShowFloatingDamageText
        self.DisableImprintDinoBuff = DisableImprintDinoBuff
        self.AllowAnyoneBabyImprintCuddle = AllowAnyoneBabyImprintCuddle
        self.OverideStructurePlatformPrevention = OverideStructurePlatformPrevention
        self.EnableExtraStructurePreventionVolumes = EnableExtraStructurePreventionVolumes
        self.NonPermanentDiseases = NonPermanentDiseases
        self.PreventDiseases = PreventDiseases
        self.OverrideStructurePlatformPrevention = OverrideStructurePlatformPrevention
        self.PreventTribeAlliances = PreventTribeAlliances
        self.AllowRaidDinoFeeding = AllowRaidDinoFeeding
        self.AllowHitMarkers = AllowHitMarkers
        self.FastDecayUnsnappedCoreStructures = FastDecayUnsnappedCoreStructures
        self.TribeLogDestroyedEnemyStructures = TribeLogDestroyedEnemyStructures
        self.OverrideOfficialDifficulty = OverrideOfficialDifficulty
        self.PreventDownloadSurvivors = PreventDownloadSurvivors
        self.PreventDownloadItems = PreventDownloadItems
        self.PreventDownloadDinos = PreventDownloadDinos
        self.PreventUploadSurvivors = PreventUploadSurvivors
        self.PreventUploadItems = PreventUploadItems
        self.PreventUploadDinos = PreventUploadDinos
        self.ForceFlyerExplosives = ForceFlyerExplosives
        self.DestroyUnconnectedWaterPipes = DestroyUnconnectedWaterPipes
        self.PvPDinoDecay = PvPDinoDecay
        self.PvEAllowStructuresAtSupplyDrops = PvEAllowStructuresAtSupplyDrops
        self.AllowCrateSpawnsOnTopOfStructures = AllowCrateSpawnsOnTopOfStructures
        self.UseOptimizedHarvestingHealth = UseOptimizedHarvestingHealth
        self.ClampItemSpoilingTimes = ClampItemSpoilingTimes
        self.AutoDestroyDecayedDinos = AutoDestroyDecayedDinos
        self.AllowFlyingStaminaRecovery = AllowFlyingStaminaRecovery
        self.AllowMultipleAttachedC4 = AllowMultipleAttachedC4
        self.bAllowPlatformSaddleMultiFloors = bAllowPlatformSaddleMultiFloors
        self.PreventSpawnAnimations = PreventSpawnAnimations
        self.AutoDestroyStructures = AutoDestroyStructures
        self.MinimumDinoReuploadInterval = MinimumDinoReuploadInterval
        self.OnlyAutoDestroyCoreStructures = OnlyAutoDestroyCoreStructures
        self.OxygenSwimSpeedStatMultiplier = OxygenSwimSpeedStatMultiplier
        self.ServerAutoForceRespawnWildDinosInterval = ServerAutoForceRespawnWildDinosInterval
        self.CrossARKAllowForeignDinoDownloads = CrossARKAllowForeignDinoDownloads
        self.ClampItemStats = ClampItemStats
        self.EnableCryopodNerf = EnableCryopodNerf
        self.NewYear1UTC = NewYear1UTC
        self.NewYear2UTC = NewYear2UTC
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




from __future__ import annotations
from ..lib.client import Client
from .log import Log


class LogsPage:
    @classmethod
    def page(cls, service_id: int, page: int) -> LogsPage:
        assert page > 0, "Page number must be greater than 0"
        path = f'/services/{service_id}/logs'
        response = Client.get(path=path, data={'page': page})
        data: dict = response.json()['data']
        data['logs'] = [Log(service_id, **log) for log in data['logs']]
        return cls(service_id, **data)

    def __init__(
            self,
            service_id: int,
            current_page: int = 1,
            logs_per_page: int = 0,
            page_count: int = 0,
            log_count: int = 0,
            logs: list = None,
            **kwargs
    ):
        self.current_page = current_page
        self.page_count = page_count
        self.log_count = log_count
        self.logs = [] if logs is None else logs
        self.logs_per_page = logs_per_page
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __iter__(self):
        return iter(self.logs)

    def __contains__(self, item):
        return item in self.logs

    def __getitem__(self, item):
        return self.logs[item]

    def __len__(self):
        return len(self.logs)

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"





class Details:
    def __init__(
            self,
            service_id: int,
            address: str = None,
            name: str = None,
            game: str = None,
            portlist_short: str = None,
            folder_short: str = None,
            slots: int = 0,
            **kwargs
    ):
        self.address = address
        self.name = name
        self.game = game
        self.portlist_short = portlist_short
        self.folder_short = folder_short
        self.slots = slots
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        name = f"name={repr(self.name)}"
        slots = f"slots={repr(self.slots)}"
        game = f"game={repr(self.game)}"
        params = [s for s in [name, slots, game]]
        return f"<{self.__class__.__name__}({', '.join(params)}, ...)>"




class General:
    @classmethod
    def from_data(cls, service_id: int, **kwargs):
        return cls(service_id, **{k.lower().replace('-', '_'): v for k, v in kwargs.items()})

    def __init__(
            self,
            service_id: int,
            expertmode: str = None,
            battleye: str = None,
            server_language: str = None,
            vday: str = None,
            mod_update_list: str = None,
            mod_status_list: str = None,
            gameplay_log: str = None,
            force_allow_cave_flyers: str = None,
            vac: str = None,
            gamesettings_saved_utc_timestamp_hidden: str = None,
            automatic_update_mechanism: str = None,
            clusterid: str = None,
            enablecluster: str = None,
            primitiveplus: str = None,
            enable_idle_player_kick: str = None,
            no_anti_speed_hack: str = None,
            no_biome_walls: str = None,
            notify_admin_commands_in_chat: str = None,
            metrics: str = None,
            crossplay: str = None,
            activeevent: str = None,
            structurememopts: str = None,
            noundermeshchecking: str = None,
            noundermeshkilling: str = None,
            newyearevent: str = None,
            useitemdupecheck: str = None,
            **kwargs
    ):
        self.expertmode = expertmode
        self.battleye = battleye
        self.server_language = server_language
        self.vday = vday
        self.mod_update_list = mod_update_list
        self.mod_status_list = mod_status_list
        self.gameplay_log = gameplay_log
        self.force_allow_cave_flyers = force_allow_cave_flyers
        self.vac = vac
        self.gamesettings_saved_utc_timestamp_hidden = gamesettings_saved_utc_timestamp_hidden
        self.automatic_update_mechanism = automatic_update_mechanism
        self.clusterid = clusterid
        self.enablecluster = enablecluster
        self.primitiveplus = primitiveplus
        self.enable_idle_player_kick = enable_idle_player_kick
        self.no_anti_speed_hack = no_anti_speed_hack
        self.no_biome_walls = no_biome_walls
        self.notify_admin_commands_in_chat = notify_admin_commands_in_chat
        self.metrics = metrics
        self.crossplay = crossplay
        self.activeevent = activeevent
        self.structurememopts = structurememopts
        self.noundermeshchecking = noundermeshchecking
        self.noundermeshkilling = noundermeshkilling
        self.newyearevent = newyearevent
        self.useitemdupecheck = useitemdupecheck
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"





class Config:
    @classmethod
    def from_data(cls, service_id: int, **kwargs):
        return cls(service_id, **{k.lower().replace('-', '_').replace('3rd', 'third'): v for k, v in kwargs.items()})

    def __init__(
            self,
            service_id: int,
            server_name: str = None,
            admin_password: str = None,
            server_password: str = None,
            player_on_map: str = None,
            disable_pvp: str = None,
            hardcore: str = None,
            crosshair: str = None,
            no_hud: str = None,
            voice_chat: str = None,
            near_chat_only: str = None,
            third_person: str = None,
            leave_messages: str = None,
            join_messages: str = None,
            message_of_the_day: str = None,
            difficulty_offset: str = None,
            motd_duration: str = None,
            disable_structure_decay_pve: str = None,
            allow_flyer_carry_pve: str = None,
            max_structures_in_range: str = None,
            enable_pvp_gamma: str = None,
            no_tribute_downloads: str = None,
            daycyclespeedscale: str = None,
            nighttimespeedscale: str = None,
            daytimespeedscale: str = None,
            dinodamagemultiplier: str = None,
            playerdamagemultiplier: str = None,
            structuredamagemultiplier: str = None,
            playerresistancemultiplier: str = None,
            dinoresistancemultiplier: str = None,
            structureresistancemultiplier: str = None,
            xpmultiplier: str = None,
            tamingspeedmultiplier: str = None,
            harvestamountmultiplier: str = None,
            harvesthealthmultiplier: str = None,
            playercharacterwaterdrainmultiplier: str = None,
            playercharacterfooddrainmultiplier: str = None,
            dinocharacterfooddrainmultiplier: str = None,
            playercharacterstaminadrainmultiplier: str = None,
            dinocharacterstaminadrainmultiplier: str = None,
            playercharacterhealthrecoverymultiplier: str = None,
            dinocharacterhealthrecoverymultiplier: str = None,
            dinocountmultiplier: str = None,
            pvestructuredecayperiodmultiplier: str = None,
            resourcesrespawnperiodmultiplier: str = None,
            clampresourceharvestdamage: str = None,
            map: str = None,
            restart_countdown_seconds: str = None,
            active_mods: str = None,
            start_with_backup: str = None,
            players_join_no_check_list: str = None,
            ban_list: str = None,
            prevent_download_survivors: str = None,
            prevent_download_items: str = None,
            prevent_download_dinos: str = None,
            admin_list: str = None,
            default_map: str = None,
            map_mod_id: str = None,
            active_total_conversion: str = None,
            disabledeathspectator: str = None,
            onlyadminrejoinasspectator: str = None,
            battlenumoftribestostartgame: str = None,
            timetocollapserod: str = None,
            battleautostartgameinterval: str = None,
            battleautorestartgameinterval: str = None,
            battlesuddendeathinterval: str = None,
            structuredestructiontag: str = None,
            forcerespawndinos: str = None,
            banlisturl: str = None,
            autosaveperiodminutes: str = None,
            activateadminlogs: str = None,
            gamelogbuffer: str = None,
            map_expert: str = None,
            disabledinodecaypve: str = None,
            pvedinodecayperiodmultiplier: str = None,
            disablepvegamma: str = None,
            exlusivejoin: str = None,
            playersexclusivejoinlist: str = None,
            forceallstructurelocking: str = None,
            autodestroyoldstructuresmultiplier: str = None,
            bjoinnotifications: str = None,
            bshowstatusnotificationmessages: str = None,
            perplatformmaxstructuresmultiplier: str = None,
            spectatorpassword: str = None,
            allowcavebuildingpve: str = None,
            nofishloot: str = None,
            disabledinoriding: str = None,
            disabledinotaming: str = None,
            maxpersonaltameddinos: str = None,
            onlydecayunsnappedcorestructures: str = None,
            tributeitemexpirationseconds: str = None,
            tributedinoexpirationseconds: str = None,
            tributecharacterexpirationseconds: str = None,
            current_admin_password: str = None,
            activateadmintribelogs: str = None,
            tribenamechangecooldown: str = None,
            allowhidedamagesourcefromlogs: str = None,
            randomsupplycratepoints: str = None,
            disableweatherfog: str = None,
            adminlogging: str = None,
            bforcecanridefliers: str = None,
            allowteksuitpowersingenesis: str = None,
            enablecryosicknesspve: str = None,
            cryopodnerfduration: str = None,
            cryopodnerfdamagemult: str = None,
            itemstacksizemultiplier: str = None,
            allowcavebuildingpvp: str = None,
            **kwargs
    ):
        self.server_name = server_name
        self.admin_password = admin_password
        self.server_password = server_password
        self.player_on_map = player_on_map
        self.disable_pvp = disable_pvp
        self.hardcore = hardcore
        self.crosshair = crosshair
        self.no_hud = no_hud
        self.voice_chat = voice_chat
        self.near_chat_only = near_chat_only
        self.third_person = third_person
        self.leave_messages = leave_messages
        self.join_messages = join_messages
        self.message_of_the_day = message_of_the_day
        self.difficulty_offset = difficulty_offset
        self.motd_duration = motd_duration
        self.disable_structure_decay_pve = disable_structure_decay_pve
        self.allow_flyer_carry_pve = allow_flyer_carry_pve
        self.max_structures_in_range = max_structures_in_range
        self.enable_pvp_gamma = enable_pvp_gamma
        self.no_tribute_downloads = no_tribute_downloads
        self.daycyclespeedscale = daycyclespeedscale
        self.nighttimespeedscale = nighttimespeedscale
        self.daytimespeedscale = daytimespeedscale
        self.dinodamagemultiplier = dinodamagemultiplier
        self.playerdamagemultiplier = playerdamagemultiplier
        self.structuredamagemultiplier = structuredamagemultiplier
        self.playerresistancemultiplier = playerresistancemultiplier
        self.dinoresistancemultiplier = dinoresistancemultiplier
        self.structureresistancemultiplier = structureresistancemultiplier
        self.xpmultiplier = xpmultiplier
        self.tamingspeedmultiplier = tamingspeedmultiplier
        self.harvestamountmultiplier = harvestamountmultiplier
        self.harvesthealthmultiplier = harvesthealthmultiplier
        self.playercharacterwaterdrainmultiplier = playercharacterwaterdrainmultiplier
        self.playercharacterfooddrainmultiplier = playercharacterfooddrainmultiplier
        self.dinocharacterfooddrainmultiplier = dinocharacterfooddrainmultiplier
        self.playercharacterstaminadrainmultiplier = playercharacterstaminadrainmultiplier
        self.dinocharacterstaminadrainmultiplier = dinocharacterstaminadrainmultiplier
        self.playercharacterhealthrecoverymultiplier = playercharacterhealthrecoverymultiplier
        self.dinocharacterhealthrecoverymultiplier = dinocharacterhealthrecoverymultiplier
        self.dinocountmultiplier = dinocountmultiplier
        self.pvestructuredecayperiodmultiplier = pvestructuredecayperiodmultiplier
        self.resourcesrespawnperiodmultiplier = resourcesrespawnperiodmultiplier
        self.clampresourceharvestdamage = clampresourceharvestdamage
        self.map = map
        self.restart_countdown_seconds = restart_countdown_seconds
        self.active_mods = active_mods
        self.start_with_backup = start_with_backup
        self.players_join_no_check_list = players_join_no_check_list
        self.ban_list = ban_list
        self.prevent_download_survivors = prevent_download_survivors
        self.prevent_download_items = prevent_download_items
        self.prevent_download_dinos = prevent_download_dinos
        self.admin_list = admin_list
        self.default_map = default_map
        self.map_mod_id = map_mod_id
        self.active_total_conversion = active_total_conversion
        self.disabledeathspectator = disabledeathspectator
        self.onlyadminrejoinasspectator = onlyadminrejoinasspectator
        self.battlenumoftribestostartgame = battlenumoftribestostartgame
        self.timetocollapserod = timetocollapserod
        self.battleautostartgameinterval = battleautostartgameinterval
        self.battleautorestartgameinterval = battleautorestartgameinterval
        self.battlesuddendeathinterval = battlesuddendeathinterval
        self.structuredestructiontag = structuredestructiontag
        self.forcerespawndinos = forcerespawndinos
        self.banlisturl = banlisturl
        self.autosaveperiodminutes = autosaveperiodminutes
        self.activateadminlogs = activateadminlogs
        self.gamelogbuffer = gamelogbuffer
        self.map_expert = map_expert
        self.disabledinodecaypve = disabledinodecaypve
        self.pvedinodecayperiodmultiplier = pvedinodecayperiodmultiplier
        self.disablepvegamma = disablepvegamma
        self.exlusivejoin = exlusivejoin
        self.playersexclusivejoinlist = playersexclusivejoinlist
        self.forceallstructurelocking = forceallstructurelocking
        self.autodestroyoldstructuresmultiplier = autodestroyoldstructuresmultiplier
        self.bjoinnotifications = bjoinnotifications
        self.bshowstatusnotificationmessages = bshowstatusnotificationmessages
        self.perplatformmaxstructuresmultiplier = perplatformmaxstructuresmultiplier
        self.spectatorpassword = spectatorpassword
        self.allowcavebuildingpve = allowcavebuildingpve
        self.nofishloot = nofishloot
        self.disabledinoriding = disabledinoriding
        self.disabledinotaming = disabledinotaming
        self.maxpersonaltameddinos = maxpersonaltameddinos
        self.onlydecayunsnappedcorestructures = onlydecayunsnappedcorestructures
        self.tributeitemexpirationseconds = tributeitemexpirationseconds
        self.tributedinoexpirationseconds = tributedinoexpirationseconds
        self.tributecharacterexpirationseconds = tributecharacterexpirationseconds
        self.current_admin_password = current_admin_password
        self.activateadmintribelogs = activateadmintribelogs
        self.tribenamechangecooldown = tribenamechangecooldown
        self.allowhidedamagesourcefromlogs = allowhidedamagesourcefromlogs
        self.randomsupplycratepoints = randomsupplycratepoints
        self.disableweatherfog = disableweatherfog
        self.adminlogging = adminlogging
        self.bforcecanridefliers = bforcecanridefliers
        self.allowteksuitpowersingenesis = allowteksuitpowersingenesis
        self.enablecryosicknesspve = enablecryosicknesspve
        self.cryopodnerfduration = cryopodnerfduration
        self.cryopodnerfdamagemult = cryopodnerfdamagemult
        self.itemstacksizemultiplier = itemstacksizemultiplier
        self.allowcavebuildingpvp = allowcavebuildingpvp
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




from __future__ import annotations
from datetime import datetime


class Token:
    @classmethod
    def from_data(cls, data: dict):
        token = data['token']
        return cls(
            id=token['id'],
            user_id=token['user']['id'],
            username=token['user']['username'],
            expires_at=token['expires_at'],
            valid_until=token['valid_until'],
            scopes=token['scopes'],
            two_factor_method=token['two_factor_method'],
            employee=token['employee'],
            valid=True,
        )

    def __init__(
            self,
            id: int = None,
            user_id: int = None,
            username: str = None,
            expires_at: int = None,
            valid_until: str = None,
            two_factor_method: str = None,
            scopes: list = None,
            employee: bool = False,
            valid: bool = False,
            **kwargs
    ):
        self.id = id
        self.user_id = user_id
        self.username = username
        self.expires_at = expires_at
        self.valid_until = None if valid_until is None else datetime.fromisoformat(valid_until)
        self.two_factor_method = two_factor_method
        self.scopes = scopes or []
        self.employee = employee
        self.valid = valid
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        username = f"username={repr(self.username)}"
        valid = f"valid={repr(self.valid)}"
        expires = f"expires_at={repr(self.expires_at)}"
        params = [s for s in [username, valid, expires]]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from __future__ import annotations
from datetime import datetime
from ..lib.client import Client


class Task:
    @classmethod
    def create_task(
            cls,
            service_id: int,
            action_method=None,
            month="*",
            day="*",
            hour="24",
            minute="0",
            weekday="*",
            action_data=None
    ) -> Task:
        path = f'/services/{service_id}/tasks'
        params = {'action_method': action_method, 'action_data': action_data, 'minute': minute,
                  'hour': hour, 'day': day, 'month': month, 'weekday': weekday}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        if not response.ok and data['status'] == 'success':
            raise Exception(response.text)
        return cls(service_id=service_id, **params)

    def __init__(
            self,
            service_id: int,
            id: int = None,
            status: str = None,
            minute: str = None,
            hour: str = None,
            day: str = None,
            month: str = None,
            weekday: str = None,
            next_run: str = None,
            last_run: str = None,
            timezone: str = None,
            action_method: str = None,
            action_data: str = None,
            **kwargs
    ):
        self.id = id
        self.status = status
        self.minute = minute
        self.hour = hour
        self.day = day
        self.month = month
        self.weekday = weekday
        self.next_run = None if next_run is None else datetime.fromisoformat(next_run)
        self.last_run = None if last_run is None else datetime.fromisoformat(last_run)
        self.timezone = timezone
        self.action_method = action_method
        self.action_data = action_data
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def update_task(self) -> bool:
        path = f'/services{self.service_id}/tasks/{self.id}'
        params = {
            'action_method': self.action_method, 'action_data': self.action_data, 'minute': self.minute,
            'hour': self.hour, 'day': self.day, 'month': self.month, 'weekday': self.weekday}
        response = Client.put(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def delete_task(self) -> bool:
        path = f'/services/{self.service_id}/tasks/{self.id}'
        response = Client.delete(path=path)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from __future__ import annotations


class Log:

    def __init__(
            self,
            service_id: int,
            category: str = None,
            severity: str = None,
            message: str = None,
            created_at: str = None,
            admin: bool = None,
            **kwargs
    ):
        self.category = category
        self.severity = severity
        self.created_at = created_at
        self.admin = admin
        self.message = message
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



class Query:
    def __init__(
            self,
            service_id: int,
            server_name: str = None,
            connect_ip: str = None,
            map: str = None,
            version: str = None,
            player_current: int = None,
            player_max: int = None,
            players: list = None,
            **kwargs
    ):
        self.server_name = server_name
        self.connect_ip = connect_ip
        self.map = map
        self.version = version
        self.player_current = player_current
        self.player_max = player_max
        self.players = players
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




class Arguments:
    def __init__(
            self,
            service_id: int,
            privacystatus: str = None,
            last_rental_time: str = None,
            startgame: str = None,
            startport: str = None,
            amount: str = None,
            limit_game_tags: str = None,
            wiuser: str = None,
            **kwargs
    ):
        self.privacystatus = privacystatus
        self.last_rental_time = last_rental_time
        self.startgame = startgame
        self.startport = startport
        self.amount = amount
        self.limit_game_tags = limit_game_tags
        self.wiuser = wiuser
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




from __future__ import annotations
from .game_features import GameFeatures


class GameSpecific:
    @classmethod
    def from_data(cls, service_id: int, **kwargs) -> GameSpecific:
        kwargs['features'] = GameFeatures(service_id, **kwargs['features'])
        return cls(service_id, **kwargs)

    def __init__(
            self,
            service_id: int,
            path: str = None,
            update_status: str = None,
            last_update: str = None,
            path_available: bool = None,
            features: GameFeatures = None,
            log_files: list = None,
            config_files: list = None,
            curseforge_customer_settings: str = None,
            **kwargs
    ):
        self.path = path
        self.update_status = update_status
        self.last_update = last_update
        self.path_available = path_available
        self.features = features
        self.log_files = log_files
        self.config_files = config_files
        self.curseforge_customer_settings = curseforge_customer_settings
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"




from __future__ import annotations
from ..lib import Client
from .details import Details
from .arguments import Arguments
from .task import Task
from .notification import Notification
from .action import Action
from .logs_page import LogsPage
from datetime import datetime


class Service:
    @classmethod
    def find_by_id(cls, service_id: int) -> Service:
        """Get a service by id."""
        response = Client.get(path=f'/services/{service_id}')
        data: dict = response.json()['data']['service']
        data['details'] = Details(service_id, **data['details'])
        data['arguments'] = Arguments(service_id, **data['arguments'])
        return cls(**data)

    @classmethod
    def all(cls) -> list[Service]:
        """Get all services."""
        response = Client.get(path='/services')
        data: dict = response.json()['data']
        servers = data['services']
        return [Service(**data) for data in servers]

    @classmethod
    def tasks_by_service_id(cls, service_id: int) -> list:
        path = f'/services/{service_id}/tasks'
        response = Client.get(path=path)
        data: dict = response.json()['data']['tasks']
        return [Task(service_id, **kwargs) for kwargs in data]

    @classmethod
    def notifications_by_service_id(cls, service_id: int) -> list:
        path = f'/services/{service_id}/notifications'
        response = Client.get(path=path)
        items: list = response.json()['data']['notifications']
        notif = []
        for data in items:
            data['actions'] = [Action(service_id, **a) for a in data['actions']]
            notif.append(Notification(**data))
        return notif

    def __init__(
            self,
            id: int = None,
            location_id: int = None,
            status: str = None,
            status_code: int = None,
            websocket_token: str = None,
            user_id: int = None,
            username: str = None,
            comment: str = None,
            auto_extension: bool = None,
            auto_extension_duration: int = None,
            auto_extension_external: bool = False,
            type: str = None,
            type_human: str = None,
            details: Details = None,
            start_date: str = None,
            suspend_date: str = None,
            delete_date: str = None,
            suspending_in: int = None,
            deleting_in: int = None,
            readonly: bool = False,
            has_benefits: bool = False,
            arguments: Arguments = None,
            roles: list = None,
            is_owner: bool = False,
            servicetype: int = None,
            **kwargs
    ):
        self.id = id
        self.location_id = location_id
        self.status = status
        self.status_code = status_code
        self.websocket_token = websocket_token
        self.user_id = user_id
        self.username = username
        self.comment = comment
        self.auto_extension = auto_extension
        self.auto_extension_duration = auto_extension_duration
        self.auto_extension_external = auto_extension_external
        self.type = type
        self.type_human = type_human
        self.details = details
        self.start_date = None if start_date is None else datetime.fromisoformat(start_date)
        self.suspend_date = None if start_date is None else datetime.fromisoformat(suspend_date)
        self.delete_date = None if start_date is None else datetime.fromisoformat(delete_date)
        self.suspending_in = suspending_in
        self.deleting_in = deleting_in
        self.readonly = readonly
        self.has_benefits = has_benefits
        self.arguments = arguments
        self.roles = roles
        self.is_owner = is_owner
        self.servicetype = servicetype
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def logs(self) -> list:
        first = LogsPage.page(self.id, 1)
        logs = first.logs
        page = 2
        while page <= first.page_count:
            data = LogsPage.page(self.id, page)
            logs += data.logs
            page += 1
        return logs

    def notifications(self) -> list:
        path = f'/services/{self.id}/notifications'
        response = Client.get(path=path)
        items: list = response.json()['data']['notifications']
        notif = []
        for data in items:
            data['actions'] = [Action(self.id, **a) for a in data['actions']]
            notif.append(Notification(**data))
        return notif

    def tasks(self) -> list:
        path = f'/services/{self.id}/tasks'
        response = Client.get(path=path)
        data: dict = response.json()['data']['tasks']
        return [Task(self.id, **kwargs) for kwargs in data]

    def __repr__(self):
        id = f"id={repr(self.id)}"
        status = f"status={repr(self.status)}"
        params = ", ".join([id, status])
        return f"<Service({params}, ...)>"




from .nitrado import initialize, gameserver_by_service_id, service_by_id, services, gameservers
from .globals import Global
from .service import Service
from .gameserver import GameServer
from .service.task import Task


__all__ = ['initialize', 'gameservers', 'gameserver_by_service_id', 'services', 'service_by_id']






from nitrado import Service
from tests.mocked_client import MockedClient as Client




def get_a_service():
    data = {
        "id": 1,
        "location_id": 3,
        "status": "active",
        "websocket_token": "41e4b1157551195144ceaf0677d2",
        "user_id": 1234567,
        "comment": None,
        "auto_extension": False,
        "auto_extension_duration": 720,
        "auto_extension_external": False,
        "type": "gameserver",
        "type_human": "Publicserver 20 Slots",
        "managedroot_id": None,
        "details": {
            "address": "5.62.66.117:10020",
            "name": "[US] - Arkpocalypse - Island",
            "game": "ARK: Survival Evolved (Xbox One)",
            "portlist_short": "arkxb",
            "folder_short": "arkxb",
            "slots": 20
        },
        "start_date": "2023-01-02T21:18:43",
        "suspend_date": "2023-04-03T05:21:17",
        "delete_date": "2023-04-10T05:21:17",
        "suspending_in": 766277,
        "deleting_in": 1371077,
        "username": "ni1234567_1",
        "roles": [
            "ROLE_OWNER"
        ]
    }
    try:
        return Service(**data)
    except Exception:
        raise Exception(f"Service could not be constructed with data: {data}")


def test_services():
    services = get_a_service()
    assert services.id == 1
    assert services.type == "gameserver"
    assert str(services.start_date) == "2023-01-02 21:18:43"
    assert str(services.suspend_date) == "2023-04-03 05:21:17"
    assert str(services.delete_date) == "2023-04-10 05:21:17"
    assert services.username == "ni1234567_1"





import os
from dotenv import load_dotenv
from .lib import Client
from .service import Service
from .gameserver import GameServer
from pathlib import Path


def initialize(api_key: str, save=True) -> None:
    """Saves the Nitrado API key in a local .env file"""
    os.environ[Client.ENV_NAME] = api_key
    if not save:
        return
    if not Path('.env').exists():
        Path('.env').touch()
    with open('.env', 'r+') as r:
        content = []
        found = False
        for line in r.readlines():
            if line.find(f'{Client.ENV_NAME}=') == 0:
                content += [f"{Client.ENV_NAME}={api_key}\n"]
                found = True
            else:
                content += [line]
        if not found:
            content = [f"{Client.ENV_NAME}={api_key}\n"] + content
        r.seek(0)
        r.write(''.join(content))
        load_dotenv()


def services() -> list[Service]:
    """Get a list of all services."""
    return Service.all()


def gameservers() -> list[GameServer]:
    """Get a list of all gameservers."""
    return GameServer.all()


def gameserver_by_service_id(service_id: int) -> GameServer:
    """Get a gameserver by service id."""
    return GameServer.find_by_id(service_id)


def service_by_id(service_id: int) -> Service:
    """Get a service by id."""
    return Service.find_by_id(service_id)





from __future__ import annotations
from ..lib import Client
from ..service import Service
from .host_systems import HostSystem
from .operating_system import OperatingSystem
from .credentials import Credentials
from .players import Players


class GameServer:
    @classmethod
    def find_by_id(cls, service_id: int) -> GameServer:
        """Get a gameserver by service id."""
        response = Client.get(path=f'/services/{service_id}/gameservers')
        data: dict = response.json()['data']['gameserver']
        data['hostsystems'] = HostSystem(service_id, **{
            name: OperatingSystem(service_id, **values)
            for name, values in data['hostsystems'].items()
        })
        data['credentials'] = Credentials(service_id, **data['credentials'])
        return GameServer(**data)

    @classmethod
    def all(cls) -> list[GameServer]:
        """Get all gameservers."""
        game_servers = []
        for service in Service.all():
            game_server = cls.find_by_id(service.id)
            game_servers.append(game_server)
        return game_servers

    def __init__(
            self,
            status: str = None,
            last_status_change: int = None,
            must_be_started: bool = None,
            websocket_token: str = None,
            hostsystems: dict = None,
            username: str = None,
            user_id: int = None,
            service_id: int = None,
            location_id: int = None,
            minecraft_mode: bool = None,
            ip: str = None,
            ipv6: str = None,
            port: int = None,
            query_port: int = None,
            rcon_port: int = None,
            label: str = None,
            type: str = None,
            memory: str = None,
            memory_mb: int = None,
            game: str = None,
            game_human: str = None,
            game_specific: dict = None,
            modpacks: dict = None,
            slots: int = None,
            location: str = None,
            credentials: dict = None,
            settings: dict = None,
            quota: str = None,
            query: dict = None,
            **kwargs
    ):
        self.status = status
        self.last_status_change = last_status_change
        self.must_be_started = must_be_started
        self.websocket_token = websocket_token
        self.hostsystems = hostsystems
        self.username = username
        self.user_id = user_id
        self.service_id = service_id
        self.location_id = location_id
        self.minecraft_mode = minecraft_mode
        self.ip = ip
        self.ipv6 = ipv6
        self.port = port
        self.query_port = query_port
        self.rcon_port = rcon_port
        self.label = label
        self.type = type
        self.memory = memory
        self.memory_mb = memory_mb
        self.game = game
        self.game_human = game_human
        self.game_specific = game_specific
        self.modpacks = modpacks
        self.slots = slots
        self.location = location
        self.credentials = credentials
        self.settings = settings
        self.quota = quota
        self.query = query
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def players(self) -> list[Players]:
        path = f'/services/{self.service_id}/gameservers/games/players'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return [Players(self.service_id, **player) for player in data['players']]

    def is_gameserver_restorable(self, folder, backup_id) -> bool:
        path = f'/services/{self.service_id}/gameservers/backups/restore_possible'
        params = {'folder': folder, 'backup': backup_id}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def gameserver_restore(self, folder, backup_id) -> bool:
        path = f'/services/{self.service_id}/gameservers/backups/gameserver'
        params = {'folder': folder, 'backup': backup_id}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def white_list_player(self, gamertag: str) -> bool:
        """
        Player_Management - Add Player to Whitelist
        :param gamertag: Player unique identifier.
        """
        params = {"identifier": gamertag}
        path = f'/services/{self.service_id}/gameservers/games/whitelist'
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def admin_list(self) -> list:
        path = f'/services/{self.service_id}/gameservers/games/adminlist'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['adminlist']

    def make_admin(self, gamertag: str) -> bool:
        """ Must whitelist the player first """
        if not self.white_list_player(gamertag):
            return False
        path = f'/services/{self.service_id}/gameservers/games/adminlist'
        params = {'identifier': gamertag}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def details(self) -> dict:
        path = f'/services/{self.service_id}/gameservers'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['gameserver']

    def download_file(self, file_path) -> dict:
        path = f'/services/{self.service_id}/gameservers/file_server/download'
        params = {'file': file_path}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data

    def rename_file(self, file_path, target_path, target_name) -> bool:
        path = f'/services/{self.service_id}/gameservers/file_server/move'
        params = {'source_path': file_path, 'target_path': target_path, 'target_filename': target_name}
        response = Client.post(path=path, params=params)
        data:dict = response.json()
        return response.ok and data['status'] == 'success'

    def start_game(self, game: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/games/start'
        params = {'game': game}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def stop_game(self, message=None, stop_message=None) -> bool:
        path = f'/services/{self.service_id}/gameservers/stop'
        params = {'message': message, 'stop_message': stop_message}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def restart_game(self, restart_message: str = None, log_message: str = None) -> bool:
        path = f'/services/{self.service_id}/gameservers/restart'
        params = {"restart_message": restart_message, "message": log_message}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def install_game(self, game: str, modpack: str = None) -> bool:
        path = f'/services/{self.service_id}/gameservers/games/install'
        params = {'game': game, 'modpack': modpack}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def uninstall_game(self, game: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/games/uninstall'
        params = {'game': game}
        response = Client.delete(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def backups_list(self) -> dict:
        path = f'/services/{self.service_id}/gameservers/backups'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['backups']

    def command(self, command) -> bool:
        path = f'/services/{self.service_id}/gameservers/app_server/command'
        params = {'command': command}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def ping(self, command) -> bool:
        path = f'/services/{self.service_id}/gameservers/app_server'
        response = Client.get(path=path, params={'command': command})
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def restore_database(self, database: str, timestamp: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/backups/database'
        params = {'database': database, 'timestamp': timestamp}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def ftp_change_password(self, password: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/ftp/password'
        params = {'password': password}
        response = Client.post(path=path, params=params)
        data:dict = response.json()
        return response.ok and data['status'] == 'success'

    def bookmarks(self) -> list:
        path = f'/services/{self.service_id}/gameservers/file_server/bookmarks'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['bookmarks']

    def file_copy(self, source_path: str, target_path: str, target_name: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/file_server/copy'
        params = {'source_path': source_path, 'target_path': target_path, 'target_name': target_name}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def create_directory(self, dir_path, name) -> bool:
        path = f'/services/{self.service_id}/gameservers/file_server/mkdir'
        params = {'path': dir_path, 'name': name}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def file_delete(self, file_path) -> bool:
        path = f'/services/{self.service_id}/gameservers/file_server/delete'
        params = {'path': file_path}
        response = Client.delete(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def files_list(self, dir_path=None, search: str = None) -> list:
        path = f'/services/{self.service_id}/gameservers/file_server/list'
        params = {'dir': dir_path, 'search': search}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data['entries']

    def file_move(self, source_path: str, target_path: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/file_server/move'
        name = source_path.split('/')[-1]
        params = {'source_path': source_path, 'target_path': target_path, 'target_file_name': name}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def file_seek(self, file_path: str, offset: int, length: int, mode: str = None) -> dict:
        path = f'/services/{self.service_id}/gameservers/file_server/seek'
        params = {'file': file_path, 'offset': offset, 'length': length, 'mode': mode}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data

    def file_size(self, file_path) -> int:
        path = f'/services/{self.service_id}/gameservers/file_server/size'
        params = {'file': file_path}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data['size']

    def files_stat(self, files_paths: list) -> list:
        path = f'/services/{self.service_id}/gameservers/file_server/stat'
        params = {'files': files_paths}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data['entries']

    def file_download(self, file_path) -> dict:
        path = f'/services/{self.service_id}/gameservers/file_server/download'
        params = {'file': file_path}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data

    def file_upload(self, target_dir: str, file_name: str) -> dict:
        path = f'/services/{self.service_id}/gameservers/file_server/upload'
        params = {'path': target_dir, 'file': file_name}
        response = Client.post(path=path, params=params)
        data: dict = response.json()['data']
        return data

    def list_games(self) -> list:
        path = f'/services/{self.service_id}/gameservers/games'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['games']

    def change_mysql_password(self, password: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/mysql/password'
        params = {'password': password}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def reset_mysql_database(self, password: str) -> bool:
        path = f'/services/{self.service_id}/gameservers/mysql/reset'
        params = {'password': password}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def package_install(self, package: str, version: str = None) -> dict:
        path = f'/services/{self.service_id}/gameservers/packages/install'
        params = {'package': package, 'version': version}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return data

    def packages_list(self) -> dict:
        path = f'/services/{self.service_id}/gameservers/packages'
        response = Client.get(path=path)
        data: dict = response.json()
        return data['packages']

    def package_reinstall(self, package: str, version: str = None) -> dict:
        path = f'/services/{self.service_id}/gameservers/packages/reinstall'
        params = {'package': package, 'version': version}
        response = Client.put(path=path, params=params)
        data: dict = response.json()
        return data

    def package_uninstall(self, package: str) -> dict:
        path = f'/services/{self.service_id}/gameservers/packages/uninstall'
        params = {'package': package}
        response = Client.delete(path=path, params=params)
        data: dict = response.json()
        return data

    def resource_usage(self, hours: int = None) -> dict:
        path = f'/services/{self.service_id}/gameservers/stats'
        params = {'hours': hours}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data['stats']

    def settings_sets_create(self, name: str = None) -> bool:
        path = f'/services/{self.service_id}/gameservers/settings/sets'
        params = {'name': name}
        response = Client.post(path=path, params=params)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def settings_sets_delete(self, set_id: int) -> bool:
        path = f'/services/{self.service_id}/gameservers/settings/sets/{set_id}'
        response = Client.delete(path=path)
        data: dict = response.json()
        return response.ok and data['success'] == 'success'

    def settings_sets(self) -> list:
        path = f'/services/{self.service_id}/gameservers/settings/sets'
        response = Client.get(path=path)
        data:dict = response.json()['data']
        return data['sets']

    def settings_sets_restore(self, set_id: int) -> bool:
        path = f'/services/{self.service_id}/gameservers/settings/sets/{set_id}/restore'
        response = Client.post(path=path)
        data: dict = response.json()
        return response.ok and data['status'] == 'success'

    def default_settings(self) -> dict:
        path = f'/services/{self.service_id}/gameservers/settings/defaults'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['settings']

    def reset_settings(self) -> bool:
        path = f'/services/{self.service_id}/gameservers/settings'
        response = Client.delete(path=path)
        data: dict = response.json()
        return response.ok and data['status'] == "success"

    def update_settings(self, category: str = None, key: str = None, value: str = None) -> dict:
        path = f'/services/{self.service_id}/gameservers/settings'
        params = {'category': category, 'key': key, 'value': value}
        response = Client.post(path=path, params=params)
        data: dict = response.json()['data']
        return data['settings']

    def boost_history(self, page: int = 1) -> dict:
        path = f'/services/{self.service_id}/gameservers/boost/history'
        params = {'page': page}
        response = Client.get(path=path, params=params)
        data: dict = response.json()['data']
        return data

    def boost_settings(self) -> dict:
        path = f'/services/{self.service_id}/gameservers/boost'
        response = Client.get(path=path)
        data: dict = response.json()['data']
        return data['boosting']

    def boost_settings_update(self, enable: bool = True, message: str = None, welcome_message: str = None) -> dict:
        path = f'/services/{self.service_id}/gameservers/boost'
        params = {'enable': enable, 'message': message, 'welcome_message': welcome_message}
        response = Client.put(path=path, params=params)
        data: dict = response.json()['data']
        return data['boosting']

    def __getitem__(self, item):
        return self.__dict__[item]

    def keys(self):
        return self.__dict__.keys()

    def __repr__(self):
        service_id = f"service_id={repr(self.service_id)}"
        status = f"status={repr(self.status)}"
        slots = f"slots={repr(self.slots)}"
        memory_mb = f"memory_mb={repr(self.memory_mb)}"
        game_human = f"game_human={repr(self.game_human)}"
        params = [s for s in [service_id, status, slots, memory_mb, game_human]]
        return f"<{self.__class__.__name__}({', '.join(params)}, ...)>"




import requests
from requests import Response


def assert_response_is_ok(response: Response):
    if not response.ok:
        code = response.status_code
        reason = response.reason
        url = response.url
        raise requests.RequestException(f"[{code} Error]: {reason} ({url})")


def assert_response_is_json(response: Response):
    headers = response.headers
    if 'Content-Type' not in headers and 'application/json' not in headers['Content-Type']:
        raise requests.RequestException(f"[JSON Error]: The response did not reply in JSON format ({response.url})")


def assert_response_is_nitrado_data(response: Response):
    data = response.json()
    if 'status' not in data:
        message = f"[Nitrado Data Error]: Response data does not match Nitrado's data standard ({response.url})"
        hint = "[Hint]: Data must be a JSON with keys containing ['status', (optional) 'message', (optional) 'data']"
        raise requests.RequestException(f"{message}\n{hint}")


def assert_success(response: Response):
    assert_response_is_ok(response)
    assert_response_is_json(response)
    assert_response_is_nitrado_data(response)
    data = response.json()

    if data['status'] != 'success':
        raise requests.RequestException(f"[Nitrado Error]: {data['message']} ({response.url})")



from __future__ import annotations
from .locations import Location


class Game:
    @classmethod
    def from_data(cls, **kwargs) -> list[Game]:
        steam_id = kwargs['steam_id']  # Not sure what this is for yet
        locations = {loc['id']: loc for loc in kwargs['locations']}
        games = []
        for game in kwargs['games']:
            game['locations'] = [Location(**locations[id]) for id in game['locations']]
            games.append(cls(**game))
        return games

    def __init__(
            self,
            id: int = None,
            steam_id: int = None,
            has_steam_game: str = None,
            name: str = None,
            minecraft_mode: bool = False,
            publicserver_only: bool = False,
            portlist_short: str = None,
            folder_short: str = None,
            minimum_slots: int = 0,
            slot_multiplier: str = None,
            maximum_recommended_slots: str = None,
            modpacks: list = None,
            icons: dict = None,
            tags: list = None,
            preorder_locations: list = None,
            orderlocked_locations: str = None,
            locations: list[int] = None,
            **kwargs
    ):
        self.id = id
        self.steam_id = steam_id
        self.has_steam_game = has_steam_game
        self.name = name
        self.minecraft_mode = minecraft_mode
        self.publicserver_only = publicserver_only
        self.portlist_short = portlist_short
        self.folder_short = folder_short
        self.minimum_slots = minimum_slots
        self.slot_multiplier = slot_multiplier
        self.maximum_recommended_slots = maximum_recommended_slots
        self.modpacks = modpacks or []
        self.icons = icons
        self.tags = tags or []
        self.preorder_locations = preorder_locations or []
        self.orderlocked_locations = orderlocked_locations
        self.locations = locations or []
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        name = f"name={repr(self.name)}"
        shorten_locations = [str(l) for l in self.locations[:3]]
        trunc_locations = ','.join(shorten_locations + (['...'] if len(self.locations) > 3 else []))
        locations = f"locations=[{trunc_locations}]"
        params = [s for s in [name, locations]]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from datetime import datetime


class Notification:
    def __init__(
            self,
            id: int = None,
            service_id: int = None,
            type: str = None,
            level: str = None,
            error_id: bool = False,
            dismissed: bool = False,
            created_at: str = None,
            lifetime: bool = False,
            message: str = None,
            message_bbcode: str = None,
            message_long: str = None,
            message_long_bbcode: str = None,
            actions: list = None,
            data: dict = None,
            created_at_timestamp: int = None,
            **kwargs
    ):
        self.id = id
        self.service_id = service_id
        self.type = type
        self.level = level
        self.error_id = error_id
        self.dismissed = dismissed
        self.created_at = None if created_at is None else datetime.fromisoformat(created_at)
        self.lifetime = lifetime
        self.message = message
        self.message_bbcode = message_bbcode
        self.message_long = message_long
        self.message_long_bbcode = message_long_bbcode
        self.actions = actions
        self.data = data
        self.created_at_timestamp = created_at_timestamp
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



class Location:
    def __init__(
            self,
            id: int = None,
            country: str = None,
            city: str = None,
            **kwargs
    ):
        self.id = id
        self.country = country
        self.city = city
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __str__(self):
        return self.country or ''

    def __repr__(self):
        country = f"country={repr(self.country)}"
        city = f"city={repr(self.city)}"
        params = [s for s in [country, city]]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from nitrado import GameServer
from tests.mocked_client import MockedClient


def get_a_game_server() -> GameServer:
    data = {
        "status": "started",
        "last_status_change": 1679636808,
        "must_be_started": True,
        "websocket_token": "415511975144cabef35bbf49eaf0e4b1157677d2",
        "hostsystems": {
            "linux": {
                "hostname": "usla652.nitrado.net",
                "servername": "usla652",
                "status": "online"
            },
            "windows": {
                "hostname": "usla992.nitrado.net",
                "servername": "usla992",
                "status": "online"
            }
        },
        "username": "ni1234567_1",
        "user_id": 123456,
        "service_id": 1,
        "location_id": 3,
        "minecraft_mode": False,
        "ip": "5.62.66.117",
        "ipv6": None,
        "port": 10020,
        "query_port": 10020,
        "rcon_port": 15335,
        "label": "ni",
        "type": "Gameserver",
        "memory": "Standard",
        "memory_mb": 1024,
        "game": "arkxb",
        "game_human": "ARK: Survival Evolved (Xbox One)",
        "game_specific": {
            "path": "/games/ni1234567_1/noftp/arkxb/",
            "update_status": "up_to_date",
            "last_update": "2023-03-06T17:19:51",
            "path_available": True,
            "features": {
                "has_backups": True,
                "has_world_backups": False,
                "has_rcon": False,
                "has_application_server": False,
                "has_container_websocket": False,
                "has_file_browser": False,
                "has_ftp": False,
                "has_expert_mode": True,
                "has_packages": False,
                "has_plugin_system": False,
                "has_restart_message_support": True,
                "has_database": False,
                "has_playermanagement_feature": True,
                "has_curseforge_workshop": False
            },
            "log_files": [
                "arkxb/ShooterGame/Saved/Logs/ShooterGame.log",
                "arkxb/ShooterGame/Saved/Logs/ShooterGame_Last.log"
            ],
            "config_files": [],
            "curseforge_customer_settings": None
        },
        "modpacks": {},
        "slots": 20,
        "location": "US",
        "credentials": {
            "ftp": None,
            "mysql": None
        },
        "settings": {
            "config": {
                "server-name": "[US] Flashy Thylacoleo hosted by nitrado.net",
                "admin-password": "QUH0TL3I",
                "server-password": "",
                "player-on-map": "False",
                "disable-pvp": "False",
                "hardcore": "False",
                "crosshair": "False",
                "no-hud": "False",
                "voice-chat": "False",
                "near-chat-only": "False",
                "3rd-person": "False",
                "leave-messages": "False",
                "join-messages": "False",
                "message-of-the-day": "This ARK is hosted by nitrado.net",
                "difficulty-offset": "0.5",
                "motd-duration": "20",
                "disable-structure-decay-pve": "False",
                "allow-flyer-carry-pve": "False",
                "max-structures-in-range": "1300",
                "enable-pvp-gamma": "False",
                "no-tribute-downloads": "False",
                "DayCycleSpeedScale": "1",
                "NightTimeSpeedScale": "1",
                "DayTimeSpeedScale": "1",
                "DinoDamageMultiplier": "1.0",
                "PlayerDamageMultiplier": "1.0",
                "StructureDamageMultiplier": "1.0",
                "PlayerResistanceMultiplier": "1.0",
                "DinoResistanceMultiplier": "1.0",
                "StructureResistanceMultiplier": "1.0",
                "XPMultiplier": "1.0",
                "TamingSpeedMultiplier": "1.0",
                "HarvestAmountMultiplier": "1.0",
                "HarvestHealthMultiplier": "1.0",
                "PlayerCharacterWaterDrainMultiplier": "1.0",
                "PlayerCharacterFoodDrainMultiplier": "1.0",
                "DinoCharacterFoodDrainMultiplier": "1.0",
                "PlayerCharacterStaminaDrainMultiplier": "1.0",
                "DinoCharacterStaminaDrainMultiplier": "1.0",
                "PlayerCharacterHealthRecoveryMultiplier": "1.0",
                "DinoCharacterHealthRecoveryMultiplier": "1.0",
                "DinoCountMultiplier": "1.0",
                "PvEStructureDecayPeriodMultiplier": "1.0",
                "ResourcesRespawnPeriodMultiplier": "1.0",
                "ClampResourceHarvestDamage": "False",
                "map": "preinstalled,1,TheIsland",
                "restart-countdown-seconds": "300",
                "active-mods": "",
                "start-with-backup": "",
                "players-join-no-check-list": "",
                "ban-list": "2533274826773268,\n2533274821865701,",
                "prevent-download-survivors": "False",
                "prevent-download-items": "False",
                "prevent-download-dinos": "False",
                "admin-list": "",
                "default-map": "True",
                "map-mod-id": "",
                "active-total-conversion": "",
                "DisableDeathSpectator": "False",
                "OnlyAdminRejoinAsSpectator": "False",
                "BattleNumOfTribesToStartGame": "15",
                "TimeToCollapseROD": "9000",
                "BattleAutoStartGameInterval": "60.000000",
                "BattleAutoRestartGameInterval": "45.000000",
                "BattleSuddenDeathInterval": "300.000000",
                "StructureDestructionTag": "False",
                "ForceRespawnDinos": "False",
                "BanListURL": "",
                "AutoSavePeriodMinutes": "30.000000",
                "activateAdminLogs": "True",
                "gameLogBuffer": "600",
                "map-expert": "",
                "DisableDinoDecayPvE": "False",
                "PvEDinoDecayPeriodMultiplier": "1.0",
                "DisablePvEGamma": "False",
                "exlusivejoin": "False",
                "PlayersExclusiveJoinList": "",
                "ForceAllStructureLocking": "False",
                "AutoDestroyOldStructuresMultiplier": "1.0",
                "bJoinNotifications": "False",
                "bShowStatusNotificationMessages": "False",
                "PerPlatformMaxStructuresMultiplier": "1.0",
                "SpectatorPassword": "",
                "AllowCaveBuildingPvE": "False",
                "nofishloot": "False",
                "DisableDinoRiding": "False",
                "DisableDinoTaming": "False",
                "MaxPersonalTamedDinos": "250",
                "OnlyDecayUnsnappedCoreStructures": "False",
                "TributeItemExpirationSeconds": "86400",
                "TributeDinoExpirationSeconds": "86400",
                "TributeCharacterExpirationSeconds": "86400",
                "current-admin-password": "564266",
                "activateAdminTribeLogs": "True",
                "TribeNameChangeCooldown": "15",
                "AllowHideDamageSourceFromLogs": "False",
                "RandomSupplyCratePoints": "False",
                "DisableWeatherFog": "True",
                "AdminLogging": "False",
                "bForceCanRideFliers": "False",
                "AllowTekSuitPowersInGenesis": "False",
                "EnableCryoSicknessPVE": "False",
                "CryopodNerfDuration": "1",
                "CryopodNerfDamageMult": "1",
                "ItemStackSizeMultiplier": "1.0",
                "AllowCaveBuildingPvP": "False"
            },
            "gameini": {
                "DinoSpawnWeightMultipliers": "",
                "OverrideEngramEntries": "",
                "LevelExperienceRampOverrides": "LevelExperienceRampOverrides=\r\nLevelExperienceRampOverrides=\r\n",
                "OverridePlayerLevelEngramPoints": "",
                "TamedDinoClassDamageMultipliers": "",
                "TamedDinoClassResistanceMultipliers": "",
                "ExcludeItemIndices": "",
                "HarvestResourceItemAmountClassMultipliers": "",
                "OverrideNamedEngramEntries": "",
                "bOnlyAllowSpecifiedEngrams": "bOnlyAllowSpecifiedEngrams=False",
                "GlobalSpoilingTimeMultiplier": "GlobalSpoilingTimeMultiplier=0",
                "GlobalItemDecompositionTimeMultiplier": "GlobalItemDecompositionTimeMultiplier=0",
                "GlobalCorpseDecompositionTimeMultiplier": "GlobalCorpseDecompositionTimeMultiplier=0",
                "OverrideMaxExperiencePointsPlayer": "",
                "OverrideMaxExperiencePointsDino": "",
                "PvPZoneStructureDamageMultiplier": "PvPZoneStructureDamageMultiplier=6.0",
                "bPvEDisableFriendlyFire": "bPvEDisableFriendlyFire=False",
                "ResourceNoReplenishRadiusPlayers": "ResourceNoReplenishRadiusPlayers=1",
                "ResourceNoReplenishRadiusStructures": "ResourceNoReplenishRadiusStructures=1",
                "bAutoPvETimer": "bAutoPvETimer=False",
                "bAutoPvEUseSystemTime": "bAutoPvEUseSystemTime=False",
                "AutoPvEStartTimeSeconds": "AutoPvEStartTimeSeconds=0",
                "AutoPvEStopTimeSeconds": "AutoPvEStopTimeSeconds=0",
                "LayEggIntervalMultiplier": "LayEggIntervalMultiplier=1.0",
                "DinoTurretDamageMultiplier": "DinoTurretDamageMultiplier=1.0",
                "bDisableLootCrates": "bDisableLootCrates=False",
                "DinoHarvestingDamageMultiplier": "DinoHarvestingDamageMultiplier=3.0",
                "bDisableFriendlyFire": "bDisableFriendlyFire=False",
                "CustomRecipeEffectivenessMultiplier": "CustomRecipeEffectivenessMultiplier=1.0",
                "CustomRecipeSkillMultiplier": "CustomRecipeSkillMultiplier=1.0",
                "MatingIntervalMultiplier": "MatingIntervalMultiplier=1.0",
                "EggHatchSpeedMultiplier": "EggHatchSpeedMultiplier=1.0",
                "BabyMatureSpeedMultiplier": "BabyMatureSpeedMultiplier=1.0",
                "bPassiveDefensesDamageRiderlessDinos": "bPassiveDefensesDamageRiderlessDinos=False",
                "KillXPMultiplier": "KillXPMultiplier=1.0",
                "HarvestXPMultiplier": "HarvestXPMultiplier=1.0",
                "CraftXPMultiplier": "CraftXPMultiplier=1.0",
                "GenericXPMultiplier": "GenericXPMultiplier=1.0",
                "SpecialXPMultiplier": "SpecialXPMultiplier=1.0",
                "PGMapName": "PGMapName=PGMap",
                "PGTerrainPropertiesString": "PGTerrainPropertiesString=",
                "ConfigOverrideSupplyCrateItems": "",
                "bDisableDinoRiding": "bDisableDinoRiding=False",
                "bDisableDinoTaming": "bDisableDinoTaming=False",
                "bUseCorpseLocator": "bUseCorpseLocator=False",
                "bDisableStructurePlacementCollision": "bDisableStructurePlacementCollision=False",
                "FastDecayInterval": "FastDecayInterval=43200",
                "bUseSingleplayerSettings": "bUseSingleplayerSettings=False",
                "bAllowUnlimitedRespecs": "bAllowUnlimitedRespecs=False",
                "SupplyCrateLootQualityMultiplier": "SupplyCrateLootQualityMultiplier=1.0",
                "FishingLootQualityMultiplier": "FishingLootQualityMultiplier=1.0",
                "PerLevelStatsMultiplier": "",
                "BabyCuddleIntervalMultiplier": "BabyCuddleIntervalMultiplier=1.0",
                "BabyCuddleGracePeriodMultiplier": "BabyCuddleGracePeriodMultiplier=1.0",
                "BabyCuddleLoseImprintQualitySpeedMultiplier": "BabyCuddleLoseImprintQualitySpeedMultiplier=1.0",
                "BabyImprintingStatScaleMultiplier": "BabyImprintingStatScaleMultiplier=1.0",
                "PlayerHarvestingDamageMultiplier": "PlayerHarvestingDamageMultiplier=1.0",
                "CropGrowthSpeedMultiplier": "CropGrowthSpeedMultiplier=1.0",
                "BabyFoodConsumptionSpeedMultiplier": "BabyFoodConsumptionSpeedMultiplier=1.0",
                "DinoClassDamageMultipliers": "",
                "bPvEAllowTribeWar": "bPvEAllowTribeWar=False",
                "bPvEAllowTribeWarCancel": "bPvEAllowTribeWarCancel=False",
                "CropDecaySpeedMultiplier": "CropDecaySpeedMultiplier=1.0",
                "HairGrowthSpeedMultiplier": "HairGrowthSpeedMultiplier=1.0",
                "FuelConsumptionIntervalMultiplier": "FuelConsumptionIntervalMultiplier=1.313000",
                "KickIdlePlayersPeriod": "KickIdlePlayersPeriod=3600",
                "MaxNumberOfPlayersInTribe": "MaxNumberOfPlayersInTribe=0",
                "UseCorpseLifeSpanMultiplier": "UseCorpseLifeSpanMultiplier=1",
                "GlobalPoweredBatteryDurabilityDecreasePerSecond": "GlobalPoweredBatteryDurabilityDecreasePerSecond=4",
                "bLimitTurretsInRange": "bLimitTurretsInRange=False",
                "LimitTurretsRange": "LimitTurretsRange=10000",
                "LimitTurretsNum": "LimitTurretsNum=100",
                "bHardLimitTurretsInRange": "bHardLimitTurretsInRange=False",
                "bShowCreativeMode": "bShowCreativeMode=False",
                "PreventOfflinePvPConnectionInvincibleInterval": "PreventOfflinePvPConnectionInvincibleInterval=5",
                "TamedDinoCharacterFoodDrainMultiplier": "TamedDinoCharacterFoodDrainMultiplier=1",
                "WildDinoCharacterFoodDrainMultiplier": "WildDinoCharacterFoodDrainMultiplier=1",
                "WildDinoTorporDrainMultiplier": "WildDinoTorporDrainMultiplier=1",
                "PassiveTameIntervalMultiplier": "PassiveTameIntervalMultiplier=1",
                "TamedDinoTorporDrainMultiplier": "TamedDinoTorporDrainMultiplier=1",
                "bIgnoreStructuresPreventionVolumes": "bIgnoreStructuresPreventionVolumes=True",
                "bGenesisUseStructuresPreventionVolumes": "False",
                "bDisableGenesisMissions": "False",
                "BabyImprintAmountMultiplier": "BabyImprintAmountMultiplier=1.0",
                "HexagonRewardMultiplier": "HexagonRewardMultiplier=1",
                "bAllowFlyerSpeedLeveling": "bAllowFlyerSpeedLeveling=False",
                "PreventTransferForClassNames": ""
            },
            "features": {
                "engine-settings": "False"
            },
            "append": {
                "gameini": ""
            },
            "general": {
                "expertMode": "True",
                "battleye": "True",
                "server-language": "en",
                "vday": "False",
                "mod-update-list": "[]",
                "mod-status-list": "{}",
                "gameplay-log": "False",
                "force-allow-cave-flyers": "True",
                "vac": "True",
                "gamesettings_saved_utc_timestamp_hidden": "1679123222",
                "automatic-update-mechanism": "AUTOMATIC_UPDATE_ON_RESTART",
                "clusterid": "CuDyawB42FxmeZoLQ+64CreomyCouLzmD+wnWNUHx7k=",
                "enablecluster": "True",
                "PrimitivePlus": "False",
                "enable-idle-player-kick": "False",
                "no-anti-speed-hack": "False",
                "no-biome-walls": "False",
                "notify-admin-commands-in-chat": "False",
                "metrics": "False",
                "CrossPlay": "True",
                "ActiveEvent": "WinterWonderland",
                "structurememopts": "True",
                "noundermeshchecking": "False",
                "noundermeshkilling": "False",
                "NewYearEvent": "True",
                "useitemdupecheck": "True"
            },
            "start-param": {
                "PvPStructureDecay": "True",
                "PreventOfflinePvP": "False",
                "PreventOfflinePvPInterval": "800",
                "ShowFloatingDamageText": "False",
                "DisableImprintDinoBuff": "False",
                "AllowAnyoneBabyImprintCuddle": "True",
                "OverideStructurePlatformPrevention": "True",
                "EnableExtraStructurePreventionVolumes": "False",
                "NonPermanentDiseases": "True",
                "PreventDiseases": "True",
                "OverrideStructurePlatformPrevention": "False",
                "PreventTribeAlliances": "False",
                "AllowRaidDinoFeeding": "True",
                "AllowHitMarkers": "True",
                "FastDecayUnsnappedCoreStructures": "True",
                "TribeLogDestroyedEnemyStructures": "True",
                "OverrideOfficialDifficulty": "0",
                "PreventDownloadSurvivors": "False",
                "PreventDownloadItems": "False",
                "PreventDownloadDinos": "False",
                "PreventUploadSurvivors": "False",
                "PreventUploadItems": "False",
                "PreventUploadDinos": "False",
                "ForceFlyerExplosives": "True",
                "DestroyUnconnectedWaterPipes": "True",
                "PvPDinoDecay": "True",
                "PvEAllowStructuresAtSupplyDrops": "False",
                "AllowCrateSpawnsOnTopOfStructures": "False",
                "UseOptimizedHarvestingHealth": "True",
                "ClampItemSpoilingTimes": "True",
                "AutoDestroyDecayedDinos": "False",
                "AllowFlyingStaminaRecovery": "False",
                "AllowMultipleAttachedC4": "False",
                "bAllowPlatformSaddleMultiFloors": "False",
                "PreventSpawnAnimations": "True",
                "AutoDestroyStructures": "False",
                "MinimumDinoReuploadInterval": "0",
                "OnlyAutoDestroyCoreStructures": "True",
                "OxygenSwimSpeedStatMultiplier": "1.0",
                "ServerAutoForceRespawnWildDinosInterval": "0",
                "CrossARKAllowForeignDinoDownloads": "False",
                "ClampItemStats": "True",
                "EnableCryopodNerf": "False",
                "NewYear1UTC": "",
                "NewYear2UTC": ""
            }
        },
        "quota": None,
        "query": {
            "server_name": "[US]- EPIX ARK -Isl1-20x/6Man/FRESHWIPE/MaxWild150/CusDrop?",
            "connect_ip": "5.62.66.117:10020",
            "map": "TheIsland",
            "version": "959.4",
            "player_current": 0,
            "player_max": 20,
            "players": []
        }
    }
    return GameServer(**data)


def test_game_server():
    game_server = get_a_game_server()
    assert game_server.service_id == 1
    assert game_server.status == 'started'
    assert game_server.user_id == 123456
    assert game_server.slots == 20
    assert game_server.query['server_name'] == '[US]- EPIX ARK -Isl1-20x/6Man/FRESHWIPE/MaxWild150/CusDrop?'




import os
from requests import get, post, put, delete, Response
from dotenv import load_dotenv, dotenv_values
from .errors import assert_success

load_dotenv()


class Client:
    ENV_NAME = "NITRADO_API_KEY"
    NITRADO_API_URL = "https://api.nitrado.net"

    @classmethod
    def headers(cls) -> dict:
        if cls.ENV_NAME in dotenv_values():
            key = dotenv_values()[cls.ENV_NAME]
        elif cls.ENV_NAME in os.environ:
            key = os.getenv(cls.ENV_NAME)
        else:
            raise Exception(f"A Nitrado API must be provided as an environment variable: {cls.ENV_NAME}")
        return {'Authorization': f'Bearer {key}'}

    @classmethod
    def make_path(cls, path: str) -> str:
        return "{}{}".format(cls.NITRADO_API_URL, path)

    @classmethod
    def get_without_api_key(cls, path: str = None, params=None, **kwargs) -> Response:
        response = get(cls.make_path(path), params=params, **kwargs)
        assert_success(response)
        return response

    @classmethod
    def get(cls, path: str = None, params=None, **kwargs) -> Response:
        response = get(cls.make_path(path), headers=cls.headers(), params=params, **kwargs)
        assert_success(response)
        return response

    @classmethod
    def post_without_api_key(cls, path: str = None, params=None, **kwargs) -> Response:
        response = post(cls.make_path(path), params=params, **kwargs)
        assert_success(response)
        return response

    @classmethod
    def post(cls, path: str = None, params=None, **kwargs) -> Response:
        response = post(cls.make_path(path), headers=cls.headers(), params=params, **kwargs)
        assert_success(response)
        return response

    @classmethod
    def delete(cls, path: str = None, params=None, **kwargs) -> Response:
        response = delete(cls.make_path(path), headers=cls.headers(), params=params, **kwargs)
        assert_success(response)
        return response

    @classmethod
    def put(cls, path: str = None, params=None, **kwargs) -> Response:
        response = put(cls.make_path(path), headers=cls.headers(), params=params, **kwargs)
        assert_success(response)
        return response



from __future__ import annotations
from .config import Config
from .feature_settings import FeatureSettings
from .general import General
from .start_param import StartParam
from .game_ini import GameIni


class Settings:
    @classmethod
    def from_data(cls, service_id: int, **kwargs) -> Settings:
        kwargs['config'] = Config.from_data(service_id, **kwargs['config'])
        kwargs['features'] = FeatureSettings(service_id, **kwargs['features'])
        kwargs['general'] = General.from_data(service_id, **kwargs['general'])
        kwargs['start_param'] = StartParam.from_data(service_id, **kwargs['start-param'])
        kwargs['gameini'] = GameIni.from_data(service_id, **kwargs['gameini'])
        return cls(service_id, **kwargs)

    def __init__(
            self,
            service_id: int,
            config: Config = None,
            gameini: GameIni = None,
            features: FeatureSettings = None,
            append: dict = None,
            general: General = None,
            start_param: StartParam = None,
            **kwargs
    ):
        self.config = config
        self.gameini = gameini
        self.features = features
        self.append = append
        self.general = general
        self.start_param = start_param
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"





class Maintenance:
    def __init__(
            self,
            cloud_backend: bool = False,
            domain_backend: bool = False,
            pmacct_backend: bool = False,
            dns_backend: bool = False
    ):
        self.cloud_backend = cloud_backend
        self.domain_backend = domain_backend
        self.pmacct_backend = pmacct_backend
        self.dns_backend: bool = dns_backend

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from .operating_system import OperatingSystem


class HostSystem:
    def __init__(
            self,
            service_id: int,
            linux: OperatingSystem = None,
            windows: OperatingSystem = None,
            macos: OperatingSystem = None,
            **kwargs
    ):
        self.linux = linux
        self.windows = windows
        self.macos = macos
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        linux = f"linux={'<OperatingSystem>' if self.linux else 'None'}"
        windows = f"windows={'<OperatingSystem>' if self.windows else 'None'}"
        macos = f"macos={'<OperatingSystem>' if self.macos else 'None'}"
        params = [s for s in [linux, windows, macos]]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



class Action:
    def __init__(self, service_id: int, title: str = None, type: str = None, url: str = None, **kwargs):
        self.title = title
        self.type = type
        self.url = url
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        params = [f'{k}={repr(v)}' for k, v in self.__dict__.items()]
        return f"<{self.__class__.__name__}({', '.join(params)})>"



from datetime import datetime


class Players:
    def __init__(
            self,
            service_id: int,
            name: str = None,
            id: str = None,
            id_type: str = None,
            online: bool = False,
            actions: list = None,
            last_online: str = None,
            **kwargs
    ):
        self.name = name
        self.id = id
        self.id_type = id_type
        self.online = online
        self.actions = actions or []
        self.last_online: datetime = None if last_online is None else datetime.fromisoformat(last_online)
        self.service_id = service_id
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        name = f"name={repr(self.name)}"
        online = f"online={repr(self.online)}"
        params = [s for s in [name, online]]
        return f"<{self.__class__.__name__}({', '.join(params)})>"
